#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
NetworkX-specific implementation of property graph abstraction
"""
from typing import Dict, Any, List, Set

import logging
import uuid
import json
import tempfile
import networkx as nx
import networkx_query as nxq

from .abc_property_graph import ABCPropertyGraph, PropertyGraphImportException, \
    PropertyGraphQueryException, ABCGraphImporter
from .networkx_mixin import NetworkXMixin


class NetworkXPropertyGraph(ABCPropertyGraph, NetworkXMixin):
    """
    This class implements ABCPropertyGraph functionality.
    It stores all graphs in a single store (similar to Neo4j) and relies
    on queries to extract the needed subgraphs. It is less efficient
    than NetworkXPropertyGraphDisjoint, however fully compliant with
    ABCPropertyGraph interface.
    """

    def __init__(self, *, graph_id: str, importer, logger=None):
        """
        :param graph_id:
        :param importer: should be an instance of NetworkXGraphImporter
        :param logger: optional
        """
        super().__init__(graph_id=graph_id, importer=importer)
        assert isinstance(importer, NetworkXGraphImporter)
        self.storage = importer.storage
        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger

    def validate_graph(self) -> None:
        """
        NetworkX offers limited opportunities for query/validation
        :return:
        """
        self._validate_all_json_properties()
        # check that all nodes and links have 'Class' property
        for n in self.storage.get_graph(self.graph_id).nodes:
            if self.storage.get_graph(self.graph_id).nodes[n].get(ABCPropertyGraph.PROP_CLASS, None) is None:
                raise PropertyGraphImportException(graph_id=self.graph_id,
                                                   msg="Some nodes are missing 'Class' property")
        for e in self.storage.get_graph(self.graph_id).edges:
            if self.storage.get_graph(self.graph_id).edges[e].get(ABCPropertyGraph.PROP_CLASS, None) is None:
                raise PropertyGraphImportException(graph_id=self.graph_id,
                                                   msg="Some edges are missing 'Class' property")

    def delete_graph(self) -> None:
        self.storage.del_graph(self.graph_id)

    def get_node_properties(self, *, node_id: str) -> (List[str], Dict[str, Any]):
        """
        get a tuple of labels (list) and properties (dict) of a node. Note that individual
        properties are strings (could be JSON encoded objects). Note that 'Class' property
        contains the equivalent of a label in Neo4j. It is not returned among the regular
        properties.
        :param node_id:
        :return: (list, dict)
        """
        assert node_id is not None
        try:
            node_props = self.storage.get_graph(self.graph_id).nodes[self._find_node(node_id=node_id)].copy()
        except KeyError:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              node_id=node_id, msg="Unable to find node")
        label = node_props.pop(self.NETWORKX_LABEL)
        return [label], node_props

    def get_node_json_property_as_object(self, *, node_id: str, prop_name: str) -> Any:
        """
        Return property as a JSON object or None if property not set. Query exception
        if property is not interpretable as JSON
        :param node_id:
        :param prop_name:
        :return:
        """
        assert node_id is not None
        assert prop_name is not None
        # very similar to Neo4j, but doesn't compare to NEO4j_NONE
        _, node_props = self.get_node_properties(node_id=node_id)
        prop_str = node_props.get(prop_name, None)
        if prop_str is None:
            return None
        try:
            prop_val = json.loads(prop_str)
            return prop_val
        except json.decoder.JSONDecodeError:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              node_id=node_id,
                                              msg=f"Unable to decode property {prop_str} as JSON.")

    def get_link_properties(self, *, node_a: str, node_b: str) -> (str, Dict[str, Any]):
        """
        get link type and properties of a link between two nodes as a tuple (no multigraphs)
        :param node_a:
        :param node_b:
        :return: (type, Dict)
        """
        assert node_a is not None
        assert node_b is not None

        # find nodes matching those IDs and get link properties
        real_node_a = self._find_node(node_id=node_a)
        real_node_b = self._find_node(node_id=node_b)
        try:
            edge_props = self.storage.get_graph(self.graph_id).edges[(real_node_a, real_node_b)].copy()
        except KeyError:
            # no edge exists
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a, msg="Link doesn't exist")
        try:
            etype = edge_props.pop(self.NETWORKX_LABEL)
            return etype, edge_props
        except KeyError:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a,
                                              msg=f"Unable to find {self.NETWORKX_LABEL} property on link")

    def update_node_property(self, *, node_id: str, prop_name: str, prop_val: Any) -> None:
        """
        update a single property of a node in a graph overwriting the previous value
        :param node_id:
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert node_id is not None
        assert prop_name is not None
        assert prop_val is not None
        # note that we are not copying node properties, which means the next line modifies
        if prop_name == self.NETWORKX_LABEL:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg=f"Changing {self.NETWORKX_LABEL} property is not permitted")
        node_props = self.storage.get_graph(self.graph_id).nodes[self._find_node(node_id=node_id)]
        node_props[prop_name] = prop_val

    def update_nodes_property(self, *, prop_name: str, prop_val: Any) -> None:
        """
        update a selected property on all nodes of the graph
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert prop_name is not None
        assert prop_val is not None
        graph_nodes = self._find_all_nodes()

        if prop_name == self.NETWORKX_LABEL:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Changing {self.NETWORKX_LABEL} property is not permitted")
        for n in graph_nodes:
            self.storage.get_graph(self.graph_id).nodes[n][prop_name] = prop_val

    def update_node_properties(self, *, node_id: str, props: Dict[str, Any]) -> None:
        """
        update multiple properties on a node (value types in dictionary must be convertible to string)
        :param node_id:
        :param props:
        :return:
        """
        assert node_id is not None
        assert props is not None
        if self.NETWORKX_LABEL in props.keys():
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg=f"Changing {self.NETWORKX_LABEL} property is not permitted")
        # gives pointer directly into properties of a node in a graph
        node_props = self.storage.get_graph(self.graph_id).nodes[self._find_node(node_id=node_id)]
        node_props.update(props)

    def update_link_property(self, *, node_a: str, node_b: str, kind: str, prop_name: str, prop_val: Any) -> None:
        """
        update a single property of a link, overriding existing property or inserting a new one
        :param node_a:
        :param node_b:
        :param kind:
        :param prop_name:
        :param prop_val:
        :return:
        """
        assert node_a is not None
        assert node_b is not None
        assert kind is not None
        assert prop_name is not None
        assert prop_val is not None

        if prop_name == self.NETWORKX_LABEL:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Changing {self.NETWORKX_LABEL} property is not permitted")
        # find nodes matching those IDs and get link properties
        real_node_a = self._find_node(node_id=node_a)
        real_node_b = self._find_node(node_id=node_b)
        try:
            # not a copy
            edge_props = self.storage.get_graph(self.graph_id).edges[(real_node_a, real_node_b)]
        except KeyError:
            # no edge exists
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a, msg="Link doesn't exist")
        etype = edge_props.get(self.NETWORKX_LABEL, None)
        if etype != kind:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a,
                                              msg="Link of this type doesn't exist")
        edge_props[prop_name] = prop_val

    def update_link_properties(self, *, node_a: str, node_b: str, kind: str, props: Dict[str, Any]) -> None:
        """
        update multiple properties of a link overriding existing ones if necessary.
        :param node_a:
        :param node_b:
        :param kind:
        :param props:
        :return:
        """
        assert node_a is not None
        assert node_b is not None
        assert kind is not None
        assert props is not None

        if self.NETWORKX_LABEL in props.keys():
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Changing {self.NETWORKX_LABEL} property is not permitted")

        real_node_a = self._find_node(node_id=node_a)
        real_node_b = self._find_node(node_id=node_b)
        try:
            # not a copy
            edge_props = self.storage.get_graph(self.graph_id).edges[(real_node_a, real_node_b)]
        except KeyError:
            # no edge exists
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a, msg="Link doesn't exist")
        etype = edge_props.get(self.NETWORKX_LABEL, None)
        if etype != kind:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_a,
                                              msg="Link of this type doesn't exist")
        edge_props.update(props)

    def list_all_node_ids(self) -> List[str]:
        """
        List all NodeID properties of nodes in a graph
        :return:
        """
        nodeids = list()
        for n in list(self._find_all_nodes()):
            nodeids.append(self.storage.get_graph(self.graph_id).nodes[n][ABCPropertyGraph.NODE_ID])
        return nodeids

    def clone_graph(self, *, new_graph_id: str):
        """
        Clone a graph to a new graph_id (only a shallow copy in NetworkX)
        :param new_graph_id:
        :return:
        """
        assert new_graph_id is not None
        new_graph = self.storage.extract_graph(self.graph_id).copy()
        self.storage.add_graph(new_graph_id, new_graph)
        return NetworkXPropertyGraph(graph_id=new_graph_id, importer=self.importer, logger=self.log)

    def serialize_graph(self) -> str:
        """
        Serialize a given graph into GraphML string or return None
        if graph is not found
        :return:
        """
        graph = self.storage.extract_graph(self.graph_id)
        graph_string = None
        if graph is not None:
            graph_string = '\n'.join(nx.generate_graphml(graph))
        return graph_string

    def graph_exists(self) -> bool:
        """
        Does the graph with this ID exist?
        :return:
        """
        query_match = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]}))
        return len(query_match) > 0

    def get_nodes_on_shortest_path(self, *, node_a: str, node_z: str, rel: str = None) -> List:
        """
        Get a list of node ids that lie on a shortest path between two nodes. Return empty
        list if no path can be found. Optionally specify the type of relationship that path
        should consist of.
        :param node_a:
        :param node_z:
        :param rel:
        :return:
        """
        # extract a graph
        graph = self.storage.extract_graph(self.graph_id)
        if graph is None:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              msg="Unable to find graph")
        # if relationship specified, drop any edge that is not of type rel from graph copy
        if rel is not None:
            self._drop_edges_not_of_type(graph, rel)
        real_node_a = self._find_node(node_id=node_a)
        real_node_z = self._find_node(node_id=node_z)
        try:
            sp = nx.shortest_path(graph, source=real_node_a, target=real_node_z)
        except nx.exception.NetworkXNoPath:
            return list()
        return self._get_node_ids_for_list(graph, sp)

    def get_first_neighbor(self, *, node_id: str, rel: str, node_label: str) -> List:
        """
        Return a list of ids of nodes of this label related via relationship. List may be empty.
        :param node_id:
        :param rel:
        :param node_label:
        :return:
        """
        assert node_id is not None
        assert rel is not None
        assert node_label is not None

        # extract a graph
        graph = self.storage.extract_graph(self.graph_id)
        if graph is None:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              msg="Unable to find graph")
        real_node = self._find_node(node_id=node_id)
        first_neighbors = set(self._get_first_neighbors_via(graph, real_node, rel))
        # filter out those neighbors that aren't of specified label
        real_first_neighbors = self._filter_nodes_by_label(graph, first_neighbors, node_label)
        # convert internal IDs into real IDs
        return self._get_node_ids_for_list(graph, real_first_neighbors)

    def get_first_and_second_neighbor(self, *, node_id: str, rel1: str, node1_label: str, rel2: str,
                                      node2_label: str) -> List:
        """
        Return a list of 2-member lists of node ids related to this node via two specified relationships.
        List may be empty.
        :param node_id:
        :param rel1:
        :param node1_label:
        :param rel2:
        :param node2_label:
        :return:
        """
        assert node_id is not None
        assert rel1 is not None
        assert node1_label is not None
        assert rel2 is not None
        assert node2_label is not None

        # extract a graph
        graph = self.storage.extract_graph(self.graph_id)
        if graph is None:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              msg="Unable to find graph nodes")
        real_node = self._find_node(node_id=node_id)
        first_neighbors = set(graph.neighbors(real_node))
        # remove first neighbors connected via relationship that isn't rel1
        neighbor_drop_list = list()
        for n in first_neighbors:
            if graph.edges[(real_node, n)].get(self.NETWORKX_LABEL, None) != rel1:
                neighbor_drop_list.append(n)

        first_neighbors = first_neighbors.difference(neighbor_drop_list)
        if len(first_neighbors) == 0:
            return list()
        # filter first neighbors by label
        first_neighbors = self._filter_nodes_by_label(graph, first_neighbors, node1_label)

        # make a list of second neighbors as neighbors of neighbors
        second_neighbors_dict = dict()
        for n in first_neighbors:
            second_neighbors = set(graph.neighbors(n))
            neighbor_drop_list = list()
            for k in second_neighbors:
                if graph.edges[(n, k)].get(self.NETWORKX_LABEL, None) != rel2:
                    neighbor_drop_list.append(n)
            second_neighbors = second_neighbors.difference(neighbor_drop_list)
            # filter second neighbors by label
            second_neighbors = self._filter_nodes_by_label(graph, second_neighbors, node2_label)
            if len(second_neighbors) > 0:
                second_neighbors_dict[n] = second_neighbors

        # convert to a list of two-member lists, converting internal IDs to guids
        ret = list()
        for k, v in second_neighbors_dict.items():
            for i in v:
                ret.append(self._get_node_ids_for_list(graph, [k, i]))

        return ret

    def delete_node(self, *, node_id: str):
        """
        Delete node from a graph (incident edges automatically deleted)
        :param node_id:
        :return:
        """
        assert node_id is not None

        self.storage.get_graph(self.graph_id).remove_node(self._find_node(node_id=node_id))

    def node_exists(self, *, node_id: str, label: str) -> bool:
        """
        Check if this node exists
        :param node_id:
        :param label:
        :return:
        """
        assert node_id is not None
        assert label is not None
        graph_nodes = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.NODE_ID, node_id]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]}
                                            ]}))
        if len(graph_nodes) > 1:
            raise PropertyGraphQueryException(node_id=node_id, graph_id=self.graph_id,
                                              msg="Duplicate node found while checking for node uniqueness")
        return len(graph_nodes) == 1

    def add_node(self, *, node_id: str, label: str, props: Dict[str, Any]) -> None:
        """
        Add a node with specified label, classes and initial properties. Properties can be empty,
        but set graph id, node id and label
        :param node_id:
        :param label:
        :param props:
        :return:
        """
        assert node_id is not None
        assert label is not None

        if self.node_exists(node_id=node_id, label=label):
            raise PropertyGraphQueryException(node_id=node_id, graph_id=self.graph_id,
                                              msg="Unable to add node - a node with this ID exists")

        int_id = self.storage.add_blank_node_to_graph(self.graph_id, Class=label,
                                                      NodeID=node_id)
        if props is not None:
            self.storage.get_graph(self.graph_id).nodes[int_id].update(props)

    def add_link(self, *, node_a: str, rel: str, node_b: str, props: Dict[str, Any] = None) -> None:
        """
        Add a link of specified types between two nodes and properties of the link
        :param node_a:
        :param rel:
        :param node_b:
        :param props:
        :return:
        """
        assert node_a is not None
        assert node_b is not None
        assert rel is not None

        real_node_a = self._find_node(node_id=node_a, graph_id=self.graph_id)
        real_node_b = self._find_node(node_id=node_b, graph_id=self.graph_id)
        if props is not None:
            self.storage.get_graph(self.graph_id).add_edge(real_node_a, real_node_b, Class=rel, **props)
        else:
            self.storage.get_graph(self.graph_id).add_edge(real_node_a, real_node_b, Class=rel)

    def find_matching_nodes(self, *, other_graph) -> Set:
        """
        Return a set of node ids that match between the two graphs
        :param other_graph: ID of the other graph
        :return:
        """
        assert other_graph is not None
        # collect NodeID properties from self and other graph as set,
        # return an intersection
        self_ids = set(self.list_all_node_ids())
        other_ids = self._collect_nodeids(self.storage.extract_graph(other_graph.graph_id))
        return self_ids.intersection(other_ids)

    def merge_nodes(self, node_id: str, other_graph, merge_properties=None):
        """
        Merge two nodes of the same id belonging to two graphs. Optionally
        specify merging behavior for individual properties. Common relationships are merged.
        Example merge properties in Python:
        {
            "name":'discard', # keep property of the caller (also if not mentioned in merged_properties)
            "age":'overwrite', # keep property of the other graph
            "kids":'combine', # make a list of properties
        }
        :param node_id:
        :param other_graph: other graph object
        :param merge_properties:
        :return:
        """
        assert node_id is not None
        assert other_graph is not None
        assert other_graph.graph_exists()

        # find out internal IDs of the two nodes in our graph and other graph
        real_node = self._find_node(node_id=node_id)
        real_other_node = self._find_node(node_id=node_id, graph_id=other_graph.graph_id)

        # save properties of both (by copy)
        node_props = self.storage.get_graph(self.graph_id).nodes[real_node].copy()
        other_props = self.storage.get_graph(other_graph.graph_id).nodes[real_other_node].copy()

        # remember that graphid is ignored in get_graph in this implementation, but respected
        # in the disjoint implementation

        # merge the nodes in situ
        nx.contracted_nodes(self.storage.get_graph(self.graph_id), real_node, real_other_node, copy=False)

        # deal with properties
        # remove all properties, including 'contracted' new property
        self.storage.get_graph(self.graph_id).nodes[real_node].clear()

        # construct a new set of properties
        new_props = dict()
        if merge_properties is None:
            new_props = node_props
        else:
            for k, v in node_props.items():
                if k in merge_properties:
                    new_props[k] = node_props[k] if merge_properties[k] == 'discard' else \
                        other_props[k] if merge_properties[k] == 'overwrite' else \
                            [node_props[k], other_props[k]] if merge_properties[k] == 'combine' else None
                else:
                    new_props[k] = node_props[k]
        self.storage.get_graph(self.graph_id).nodes[real_node].update(new_props)


class NetworkXGraphStorage:
    """
    Shell for singleton storing all graphs in-memory. Graphs
    are stored in a single NetworkX graph object. Care is taken
    when loading new graphs to disambiguate their vertices.
    For the moment the code is not thread-safe.
    """

    class __NetworkXGraphStorage:
        """
        Singleton in-memory storage for graphs
        """

        def __init__(self):
            self.graphs = nx.Graph()
            self.start_id = 1

        def add_graph(self, graph_id: str, graph: nx.Graph) -> None:
            # relabel incoming graph nodes to integers, then merge
            temp_graph = nx.convert_node_labels_to_integers(graph, first_label=self.start_id)
            # set/overwrite GraphID property on all nodes
            for n in list(temp_graph.nodes()):
                if not temp_graph.nodes[n].get(ABCPropertyGraph.NODE_ID, None):
                    raise PropertyGraphImportException(graph_id=graph_id,
                                                       msg="Some nodes are missing NodeID property, unable to import")
                temp_graph.nodes[n][ABCPropertyGraph.GRAPH_ID] = graph_id
            self.start_id = self.start_id + len(temp_graph.nodes())
            self.graphs.add_nodes_from(temp_graph.nodes(data=True))
            self.graphs.add_edges_from(temp_graph.edges(data=True))

        def add_graph_direct(self, graph_id: str, graph: nx.Graph) -> None:
            # relabel incoming graph nodes to integers, then merge
            temp_graph = nx.convert_node_labels_to_integers(graph, first_label=self.start_id)
            self.start_id = self.start_id + len(temp_graph.nodes())
            self.graphs.add_nodes_from(temp_graph.nodes(data=True))
            self.graphs.add_edges_from(temp_graph.edges(data=True))

        def del_graph(self, graph_id: str) -> None:
            # find all nodes of this graph and remove them
            graph_nodes = list(nxq.search_nodes(self.graphs, {'eq': [ABCPropertyGraph.GRAPH_ID, graph_id]}))
            if graph_nodes is not None and len(graph_nodes) > 0:
                self.graphs.remove_nodes_from(graph_nodes)

        def extract_graph(self, graph_id: str) -> nx.Graph or None:
            # extract copy of graph from store or return None
            graph_nodes = list(nxq.search_nodes(self.graphs, {'eq': [ABCPropertyGraph.GRAPH_ID, graph_id]}))
            if len(graph_nodes) == 0:
                return None
            # get adjacency (only gets edges and their properties)
            edge_dict = nx.to_dict_of_dicts(self.graphs, graph_nodes)
            # create new graph from edges
            ret = nx.from_dict_of_dicts(edge_dict)
            for n in graph_nodes:
                # merge node dictionaries
                ret.nodes[n].update(self.graphs.nodes[n])
            return ret

        def get_graph(self, graph_id) -> nx.Graph:
            # return the store for all graphs (graph_id is ignored)
            return self.graphs

        def del_all_graphs(self) -> None:
            self.graphs.clear()

        def add_blank_node_to_graph(self, graph_id, **attrs) -> int:
            # add a new node into a graph, return internal
            # int id of the added node
            self.graphs.add_node(self.start_id, GraphID=graph_id, **attrs)
            self.start_id = self.start_id + 1
            return self.start_id - 1

    storage_instance = None

    def __init__(self):
        if not NetworkXGraphStorage.storage_instance:
            NetworkXGraphStorage.storage_instance = NetworkXGraphStorage.__NetworkXGraphStorage()

    def __getattr__(self, name):
        return getattr(self.storage_instance, name)


class NetworkXGraphImporter(ABCGraphImporter):
    """
    Importer for NetworkX graphs. Stores graphs in a single NetworkX
    object. Care is taken to disambiguate nodes when loading graphs.
    """

    def __init__(self, *, logger=None):
        """
        Initialize the importer setting up storage and logger
        :param logger:
        """
        self.storage = NetworkXGraphStorage()
        self.graph_class = NetworkXPropertyGraph
        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger

    @staticmethod
    def _relabel_nodes_to_nodeids(graph: nx.Graph, copy: bool) -> nx.Graph:
        """
        Relabel node names to their node ids. If copy set to True, produce a
        copy, otherwise do it in place. Note that normally nodes are renamed
        to have integer IDs on import.
        :param graph:
        :param copy:
        :return:
        """
        name_map = dict()
        for n in list(graph.nodes):
            name_map[n] = graph.nodes[n][ABCPropertyGraph.NODE_ID]

        # not every node has a NodeID property
        if len(graph.nodes) != len(name_map):
            raise PropertyGraphImportException(graph_id=None,
                                               msg="Some nodes are missing NodeID property, unable to relabel")

        return nx.relabel_nodes(graph, name_map, copy)

    def import_graph_from_string(self, *, graph_string: str, graph_id: str = None) -> ABCPropertyGraph:
        """
        Load graph serialized as a GraphML string. Assign graph id property to nodes.
        :param graph_string:
        :param graph_id:
        :return:
        """
        assert graph_string is not None

        if graph_id is None:
            graph_id = str(uuid.uuid4())

        with tempfile.NamedTemporaryFile(suffix="-graphml", mode='w') as f1:
            f1.write(graph_string)
            # read using networkx
            self.storage.add_graph(graph_id=graph_id, graph=nx.read_graphml(f1.name))

        return self.graph_class(graph_id=graph_id, importer=self, logger=self.log)

    def import_graph_from_string_direct(self, *, graph_string: str) -> ABCPropertyGraph:
        """
        Load graph serialized as a GraphML string with no checks or manipulations. It is
        assumed the GraphML already contains graph ids and node ids.
        :param graph_string:
        :return:
        """
        assert graph_string is not None

        with tempfile.NamedTemporaryFile(suffix="-graphml", mode='w') as f1:
            f1.write(graph_string)
            # get graph id (kinda inefficient, because reads graph in, then discards)
            graph_id = self.get_graph_id(graph_file=f1.name)
            # read using networkx
            self.storage.add_graph_direct(graph_id=graph_id, graph=nx.read_graphml(f1.name))

        return self.graph_class(graph_id=graph_id, importer=self, logger=self.log) if graph_id is not None else None

    def import_graph_from_file_direct(self, *, graph_file: str) -> ABCPropertyGraph:
        """
        Import a graph from file without any manipulations
        :param graph_file: name of the file
        :return:
        """
        assert graph_file is not None
        # get graph id
        graph_id = self.get_graph_id(graph_file=graph_file)
        self.storage.add_graph_direct(graph_id=graph_id, graph=nx.read_graphml(graph_file))
        return self.graph_class(graph_id=graph_id, importer=self, logger=self.log) if graph_id is not None else None

    def delete_graph(self, *, graph_id: str) -> None:
        """
        Delete a single graph with this id
        :param graph_id:
        :return:
        """
        assert graph_id is not None
        self.storage.del_graph(graph_id)

    def delete_all_graphs(self) -> None:
        """
        Remove all graphs from the by-GUID in-memory dictionary
        :return:
        """
        self.storage.del_all_graphs()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Neo4j implementation of CBM (Combined Broker Model) functionality.
"""
import uuid
import json

from typing import List

from ..neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter
from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from .abc_cbm import ABCCBMMixin
from .neo4j_adm import Neo4jADMGraph
from fim.slivers.delegations import DelegationType

from ...pluggable import PluggableRegistry, BrokerPluggable, PluggableType


class Neo4jCBMGraph(Neo4jPropertyGraph, ABCCBMMixin):
    """
    Neo4j implementation of CBM
    """
    ADM_GRAPH_IDS = "ADMGraphIDs"
    BQM_MERGED_FIELDS = ['LabelDelegations', 'CapacityDelegations']
    DELEGATION_TYPE_TO_PROP_NAME = {DelegationType.LABEL: ABCPropertyGraph.PROP_LABEL_DELEGATIONS,
                                    DelegationType.CAPACITY: ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS
                                    }

    def __init__(self, *, graph_id: str = None, importer: Neo4jGraphImporter, logger=None):
        """
        Initialize CBM either from existing ID or generate new graph id
        :param graph_id:
        :param importer:
        """
        if graph_id is None:
            graph_id = str(uuid.uuid4())

        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def _update_node_delegations(self, *, node_id: str, adm: ABCPropertyGraph) -> None:
        """
        Update the CBM node by-ADM-id delegation dictionary (it is assumed to exist already
        on common nodes) from corresponding ADM node.
        Must be invoked before ADM is merged into CBM.
        :param node_id:
        :param adm_id:
        :return:
        """
        _, cbm_node_props = self.get_node_properties(node_id=node_id)
        _, adm_node_props = adm.get_node_properties(node_id=node_id)
        props_modified = False
        for delegation_prop_name in [self.PROP_LABEL_DELEGATIONS, self.PROP_CAPACITY_DELEGATIONS]:
            # one or both of properties can have data
            if delegation_prop_name not in cbm_node_props.keys() or \
                    cbm_node_props[delegation_prop_name] == self.NEO4j_NONE:
                cbm_delegation_dict = dict()
            else:
                cbm_delegation_dict = json.loads(cbm_node_props[delegation_prop_name])

            if delegation_prop_name not in adm_node_props.keys() or \
                    adm_node_props[delegation_prop_name] == self.NEO4j_NONE:
                continue
            props_modified = True
            # turn delegation into Python object by merging two dictionaries
            adm_delegation = json.loads(adm_node_props[delegation_prop_name])
            cbm_delegation_dict.update(adm_delegation)
            new_delegation_json = json.dumps(cbm_delegation_dict)
            cbm_node_props[delegation_prop_name] = new_delegation_json
        if props_modified:
            # write back
            self.update_node_properties(node_id=node_id, props=cbm_node_props)

    def merge_adm(self, *, adm: ABCPropertyGraph) -> None:
        """
        Merge an ADM into the CBM
        :param adm:
        :return:
        """
        assert adm is not None
        assert adm.graph_exists()

        # clone ADM with temporary ID - we will be changing its properties
        temp_adm_graph_id = str(uuid.uuid4())
        temp_graph = adm.clone_graph(new_graph_id=temp_adm_graph_id)
        # crude typecasting
        temp_adm_graph = Neo4jADMGraph(graph_id=temp_graph.graph_id,
                                       importer=temp_graph.importer,
                                       logger=temp_graph.log)

        # rewrite delegations into dictionaries keyed by ADM graph id,
        # but be sure to use original ADM graph ID when rewriting
        temp_adm_graph.rewrite_delegations(real_adm_id=adm.graph_id)

        # update the ADMGraphIDs property to a single member list on all CBM nodes
        temp_adm_graph.update_nodes_property(prop_name=self.ADM_GRAPH_IDS,
                                             prop_val=json.dumps([adm.graph_id]))

        # if CBM is empty, just force ADM into it
        if not self.graph_exists():
            temp_adm_graph.update_nodes_property(prop_name="GraphID", prop_val=self.graph_id)
            return

        # CBM is not empty - need to merge ADM with it

        # locate nodes with matching IDs
        common_node_ids = self.find_matching_nodes(other_graph=temp_adm_graph)

        # Merging CBM properties with ADM (on merged nodes):
        # edited after the fact.
        # GraphID: use CBM
        # NodeID: use CBM (should be same anyway)
        # Labels: use CBM - may have information from reservations already
        # Capacities: use CBM - may have information from reservations already - should be same
        # in both graphs initially
        # LabelDelegations: use CBM, edit/add from ADM
        # CapacityDelegations: use CBM, edit/add from ADM
        for node_id in common_node_ids:
            # update Label and Capacity delegation (now they are dictionaries, not lists)
            # prior to merge
            self._update_node_delegations(node_id=node_id, adm=temp_adm_graph)

            # default behavior - keep CBM node properties, discard
            # temp_adm_graph properties
            self.merge_nodes(node_id=node_id,
                             other_graph=temp_adm_graph)

            # add delegation graph id to ADMGraphIDs property on common (merged) nodes
            _, cbm_node_props = self.get_node_properties(node_id=node_id)
            adm_list = json.loads(cbm_node_props[self.ADM_GRAPH_IDS])
            adm_list.append(adm.graph_id)
            self.update_node_property(node_id=node_id, prop_name=self.ADM_GRAPH_IDS,
                                      prop_val=json.dumps(adm_list))

        # rewrite GraphID on the remaining nodes of
        # temporary ADM graph (after that it ceases to exist)
        # NOTE: this takes advantage of Neo4j semantics of common store for all graphs
        # and changing the GraphID property effectively makes graph takes on a new identity
        temp_adm_graph.update_nodes_property(prop_name="GraphID", prop_val=self.graph_id)

    def get_bqm(self, **kwargs) -> Neo4jPropertyGraph:
        """
        Get a Broker Query Model (e.g. in response to a listResources() call)
        :param kwargs:
        :return:
        """
        p = PluggableRegistry()
        if p.pluggable_registered(t=PluggableType.Broker) and \
                BrokerPluggable.PLUGGABLE_PRODUCE_BQM in p.get_implemented_methods(t=PluggableType.Broker):
            # invoke the callable for plug_produce_bqm() otherwise default behavior
            c = p.get_method_callable(t=PluggableType.Broker,
                                      method=BrokerPluggable.PLUGGABLE_PRODUCE_BQM)
            return c(cbm=self, **kwargs)
        # default for now - simplest possible - return a cloned CBM

        bqm = self.clone_graph(new_graph_id=str(uuid.uuid4()))
        return bqm

    def get_delegations(self, *, node_id: str, adm_id: str, delegation_type: DelegationType) -> List:
        """
        Retrieve Label or Capacity delegations attribuited to a particular ADM. A delegation property
        is a dictionary of per ADM graph id delegations. Individual delegations
        are lists of dictionaries with possible pool designators. Returns None if delegation property is
        not defined at all or delegation dictionary doesn't contain ADM graph id.
        :param node_id:
        :param adm_id:
        :param delegation_type:
        :return:
        """
        assert node_id is not None
        assert adm_id is not None
        prop_field_name = self.DELEGATION_TYPE_TO_PROP_NAME[delegation_type]
        deleg_prop_val = self.get_node_json_property_as_object(node_id=node_id, prop_name=prop_field_name)
        if deleg_prop_val is None:
            return None
        if not isinstance(deleg_prop_val, dict):
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg=f"Expected CBM {delegation_type.name} "
                                                  f"delegation to be a dictionary, got {deleg_prop_val} "
                                                  f"of {type(deleg_prop_val)}")
        #print(f"Looking for {delegation_type} delegation from {adm_id} in {deleg_prop_val}")
        return deleg_prop_val.get(adm_id, None)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from abc import ABC, abstractmethod
from typing import List, Dict, Any
import enum

import uuid
import networkx as nx
import matplotlib.pyplot as plt

from ..slivers.network_node import NodeType, NodeSliver
from ..slivers.network_link import LinkType, NetworkLinkSliver
from ..slivers.switch_fabric import SFLayer
from ..graph.slices.networkx_asm import NetworkxASM
from ..graph.networkx_property_graph import NetworkXGraphImporter
from ..graph.abc_property_graph import ABCPropertyGraph

from .model_element import ElementType
from .node import Node
from .interface import Interface
from .link import Link


class TopologyDetail(enum.Enum):
    """
    Describe the level of detail for drawing a topology.
    """
    Derived = enum.auto()
    WithComponents = enum.auto()
    InfoModel = enum.auto()


class Topology(ABC):
    """
    Define and manipulate a topology over its life cycle
    """
    def __init__(self, logger=None):
        self.graph_model = NetworkxASM(graph_id=str(uuid.uuid4()),
                                       importer=NetworkXGraphImporter(logger=logger),
                                       logger=logger)

    def add_node(self, *, name: str, node_id: str = None, site: str, ntype: NodeType = NodeType.VM,
                 **kwargs) -> Node:
        """
        Add a network node into the substrate topology. Be sure to hardcode node_id - it should
        be consistent from revision to revision of the topology for the same objects.
        kwargs serve to add parameters to the node
        cpu_cores
        ram_size
        disk_size
        image_type
        image_ref
        :param name:
        :param node_id:
        :param site:
        :param ntype:
        :return:
        """
        assert name is not None
        assert site is not None
        # add node to graph
        n = Node(name=name, node_id=node_id, topo=self, site=site,
                 etype=ElementType.NEW, ntype=ntype, **kwargs)
        return n

    def remove_node(self, name: str):
        """
        Remove node and all its components, its interfaces and interfaces of components
        :param name:
        :return:
        """
        assert name is not None
        self.graph_model.remove_network_node_with_components_sfs_cps_and_links(
            node_id=self.__get_node_by_name(name=name).node_id)

    def add_link(self, *, name: str, node_id: str = None, ltype: LinkType = None,
                 interfaces: List[Interface], layer: SFLayer = None, technology: str = None,
                 **kwargs) -> Link:
        """
        Add link between listed interfaces with specified parameters
        :param name:
        :param node_id:
        :param ltype:
        :param interfaces:
        :param layer:
        :param technology:
        :param kwargs:
        :return:
        """
        # add link to graph
        link = Link(name=name, node_id=node_id, ltype=ltype, interfaces=interfaces, **kwargs,
                    etype=ElementType.NEW, topo=self, layer=layer, technology=technology)
        return link

    def remove_link(self, name: str):
        """
        Remove a link between interfaces
        :param name:
        :return:
        """
        assert name is not None
        self.graph_model.remove_network_link(node_id=self.__get_link_by_name(name=name).node_id)

    def __get_node_by_name(self, name: str) -> Node:
        """
        Find node by its name, return Node object
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.graph_model.find_node_by_name(node_name=name,
                                                     label=ABCPropertyGraph.CLASS_NetworkNode)
        return Node(name=name, node_id=node_id, topo=self)

    def __get_link_by_name(self, name: str) -> Link:
        """
        Find link by its name, return Link object
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.graph_model.find_node_by_name(node_name=name,
                                                     label=ABCPropertyGraph.CLASS_Link)
        return Link(name=name, node_id=node_id, topo=self)

    def __get_node_by_id(self, node_id: str) -> Node:
        """
        Get node by its node_id, return Node object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Node(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id, topo=self)

    def __get_link_by_id(self, node_id: str) -> Link:
        """
        Get link by its node_id, return Link object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Link(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id, topo=self)

    def __list_nodes(self) -> Dict[str, Node]:
        """
        List all NetworkNodes in the topology as a dictionary
        organized by node name. Modifying the dictionary will not affect
        the underlying model, but modifying Nodes in the dictionary will.
        :return:
        """
        node_id_list = self.graph_model.get_all_network_nodes()
        # Could consider using frozendict or other immutable idioms
        ret = dict()
        for nid in node_id_list:
            n = self.__get_node_by_id(nid)
            ret[n.name] = n
        return ret

    def __list_links(self) -> Dict[str, Link]:
        """
        List all Links in the topology as a dictionary organized by Link name.
        :return:
        """
        link_id_list = self.graph_model.get_all_network_links()
        ret = dict()
        for nid in link_id_list:
            n = self.__get_link_by_id(nid)
            ret[n.name] = n
        return ret

    def __list_interfaces(self) -> Dict[str, Interface]:
        """
        List all interfaces of the topology as a dictionary
        :return:
        """
        ret = dict()
        for n in self.nodes.values():
            ret.update(n.interfaces)
        return ret

    def __getattr__(self, item):
        """
        Special handling for attributes like 'nodes' and 'links' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """
        if item == 'nodes':
            return self.__list_nodes()
        if item == 'links':
            return self.__list_links()
        if item == 'interfaces':
            return self.__list_interfaces()

    def serialize(self, file_name=None) -> Any:
        """
        Serialize to string or to file, depending on whether file_name
        is provided.
        :param file_name:
        :return string containing GraphML if file_name is None:
        """
        graph_string = self.graph_model.serialize_graph()
        if file_name is None:
            return graph_string
        else:
            with open(file_name, 'w') as f:
                f.write(graph_string)
        return None

    def draw(self, *, file_name: str = None, interactive: bool = False,
             topo_detail: TopologyDetail = TopologyDetail.InfoModel,
             layout=nx.spring_layout):
        """
        Use pyplot to draw the derived topology
        :param file_name: save figure to a file (drawing type is determined by extension, e.g. .png)
        :param interactive: use interactive pyplot mode (defaults to False)
        :param topo_detail: level of detail to use in drawing, defaults to OnlyNodes
        :param layout: use one of available layout algorithms in NetworkX (defaults to spring_layout)
        :return:
        """
        # clear the drawing
        if not interactive:
            plt.ioff()
        else:
            plt.ion()
        plt.clf()
        if topo_detail == TopologyDetail.InfoModel:
            # create dictionaries of names for nodes and labels for edges
            g = self.graph_model.storage.extract_graph(self.graph_model.graph_id)
            if g is None:
                return
            node_labels = dict()
            for n in g.nodes:
                node_labels[n] = g.nodes[n][ABCPropertyGraph.PROP_NAME] + \
                                 '[' + g.nodes[n][ABCPropertyGraph.PROP_TYPE] + ']'
            edge_labels = dict()
            for e in g.edges:
                edge_labels[e] = g.edges[e][ABCPropertyGraph.PROP_CLASS]
            # run the layout
            pos = layout(g)
            # draw
            nx.draw_networkx(g, labels=node_labels, pos=pos)
            nx.draw_networkx_edge_labels(g, edge_labels=edge_labels, pos=pos)
            if not interactive:
                plt.show()
            if file_name is not None:
                plt.savefig(file_name)
        elif topo_detail == TopologyDetail.Derived:
            # collect all network nodes and links, draw edges between them
            network_nodes = self.graph_model.get_all_network_nodes()
            links = self.graph_model.get_all_network_links()

            derived_graph = nx.Graph()
            for n in network_nodes:
                _, props = self.graph_model.get_node_properties(node_id=n)
                derived_graph.add_node(props[ABCPropertyGraph.PROP_NAME])
            for n in links:
                _, props = self.graph_model.get_node_properties(node_id=n)
                derived_graph.add_node(props[ABCPropertyGraph.PROP_NAME])

            # add edges into the graph
            # FIXME
            pos = layout(derived_graph)
            nx.draw_networkx(derived_graph, pos=pos)
            if not interactive:
                plt.show()
            if file_name is not None:
                plt.savefig(file_name)
        else:
            raise RuntimeError("This level of detail not yet implemented")


class ExperimentTopology(Topology):
    """
    Define an user topology model
    """
    def __init__(self, logger=None):
        super().__init__(logger)


class SubstrateTopology(Topology):
    """
    Define an substrate topology model.
    """
    def __init__(self, logger=None):
        super().__init__(logger)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
NetworkX-specific implementation of property graph abstraction
"""
from typing import Dict, Any, Tuple, List, Set

import logging
import tempfile
import uuid
import networkx as nx

from .abc_property_graph import ABCPropertyGraph, PropertyGraphImportException, PropertyGraphQueryException
from .networkx_property_graph import NetworkXPropertyGraph, NetworkXGraphImporter


class NetworkXPropertyGraphDisjoint(NetworkXPropertyGraph):
    """
    This class implements most of ABCPropertyGraph functionality.
    It stores each graph as a separate object and as a result is more
    efficient, however cannot implement merge_nodes properly - a RuntimeError
    will be raised if the method is called.
    For a fully semantics-compliant implementation please see
    NetworkXPropertyGraph.
    """

    def __init__(self, *, graph_id: str, importer, logger=None):
        # don't call immediate predecessor
        super().__init__(graph_id=graph_id, importer=importer)
        assert isinstance(importer, NetworkXGraphImporterDisjoint)

    def merge_nodes(self, node_id: str, other_graph, merge_properties=None):
        """
        Not implementable in NetworkX with graphs stored separately -
        has assumptions about a common store for all graphs.
        Would require to store all graphs in a single NetworkX graph.
        """
        raise RuntimeError("Not implementable")

    def graph_exists(self) -> bool:
        """
        Does the graph with this ID exist?
        :return:
        """
        return self.storage.graphs.get(self.graph_id, None) is not None


class NetworkXGraphStorageDisjoint:
    """
    Shell for singleton storing all graphs in-memory. Graphs
    are stored as separate NetworkX objects.
    """

    class __NetworkXGraphStorage:
        """
        Singleton in-memory storage for graphs stored separately by id
        """

        def __init__(self):
            self.graphs = dict()

        def add_graph(self, graph_id: str, graph: nx.Graph) -> None:
            # relabel incoming graph nodes to integers, then add
            temp_graph = nx.convert_node_labels_to_integers(graph, 1)
            # set/overwrite GraphID property on all nodes
            for n in list(temp_graph.nodes()):
                if not temp_graph.nodes[n].get(ABCPropertyGraph.NODE_ID, None):
                    raise PropertyGraphImportException(graph_id=graph_id,
                                                       msg="Some nodes are missing NodeID property, unable to import")
                temp_graph.nodes[n][ABCPropertyGraph.GRAPH_ID] = graph_id
            # this is needed in case passed in graph is actually a DiGraph
            # which graphs produced by yEd tend to be
            self.graphs[graph_id] = nx.Graph()
            self.graphs[graph_id].add_nodes_from(temp_graph.nodes(data=True))
            self.graphs[graph_id].add_edges_from(temp_graph.edges(data=True))

        def add_graph_direct(self, graph_id: str, graph: nx.Graph) -> None:
            # relabel incoming graph nodes to integers, then merge
            temp_graph = nx.convert_node_labels_to_integers(graph, 1)
            self.graphs[graph_id] = temp_graph

        def del_graph(self, graph_id: str) -> None:
            try:
                self.graphs.pop(graph_id)
            except KeyError:
                pass

        def extract_graph(self, graph_id: str) -> nx.Graph or None:
            graph = self.graphs.get(graph_id, None)
            if graph is not None:
                return graph.copy()
            else:
                return None

        def get_graph(self, graph_id) -> nx.Graph:
            # return the store for this graph
            return self.graphs.get(graph_id, None)

        def del_all_graphs(self) -> None:
            self.graphs.clear()

        def add_blank_node_to_graph(self, graph_id, **attrs) -> int:
            # add a new node into a graph, return internal
            # int id of the added node
            new_id = 1 + len(self.graphs[graph_id].nodes())
            self.graphs[graph_id].add_node(new_id, GraphID=graph_id, **attrs)
            return new_id

    storage_instance = None

    def __init__(self):
        if not NetworkXGraphStorageDisjoint.storage_instance:
            NetworkXGraphStorageDisjoint.storage_instance = NetworkXGraphStorageDisjoint.__NetworkXGraphStorage()

    def __getattr__(self, name):
        return getattr(self.storage_instance, name)


class NetworkXGraphImporterDisjoint(NetworkXGraphImporter):
    """
    Importer for NetworkX graphs. Stores graphs separately a dictionary
    based on GUID strings.
    """

    def __init__(self, *, logger=None):
        """
        Initialize the importer setting up storage and logger
        :param logger:
        """
        self.storage = NetworkXGraphStorageDisjoint()
        self.graph_class = NetworkXPropertyGraphDisjoint
        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
"""
Base class for all sliver types
"""
from typing import Any, List
from abc import ABC, abstractmethod

from fim.slivers.capacities_labels import Capacities, Labels


class BaseSliver(ABC):
    """Base class for all sliver types"""

    @abstractmethod
    def __init__(self):
        self.resource_type = None
        self.resource_name = None
        self.resource_model = None
        self.capacities = None
        self.labels = None
        self.node_id = None
        self.details = None

    def set_resource_type(self, resource_type):
        """
        Could be NodeType, ComponentType, InterfaceType etc
        :param resource_type:
        :return:
        """
        self.resource_type = resource_type

    def get_resource_type(self):
        return self.resource_type

    def get_type(self):
        """
        Aliased to get_resource_type
        :return:
        """
        return self.get_resource_type()

    def set_resource_name(self, resource_name: str):
        self.resource_name = resource_name

    def set_name(self, resource_name: str):
        self.resource_name = resource_name

    def get_resource_name(self):
        return self.resource_name

    def get_name(self):
        return self.resource_name

    def set_resource_model(self, resource_model: str):
        self.resource_model = resource_model

    def set_model(self, resource_model: str):
        """
        Aliased to set_resource_model
        :param resource_model:
        :return:
        """
        self.resource_model = resource_model

    def get_resource_model(self):
        return self.resource_model

    def get_model(self):
        return self.resource_model

    def set_capacities(self, cap: Capacities) -> None:
        self.capacities = cap

    def get_capacities(self) -> Capacities:
        return self.capacities

    def set_labels(self, lab: Labels) -> None:
        self.labels = lab

    def get_labels(self) -> Labels:
        return self.labels

    def set_details(self, desc: str) -> None:
        self.details = desc

    def get_details(self) -> str:
        return self.details

    def set_properties(self, **kwargs):
        """
        Lets you set multiple properties exposed via setter methods
        :param kwargs:
        :return:
        """
        # set any property on a sliver that has a setter
        for k, v in kwargs.items():
            try:
                # we can set anything the sliver model has a setter for
                self.__getattribute__('set_' + k)(v)
            except AttributeError:
                raise RuntimeError(f'Unable to set property {k} on the sliver - no such property available')

    @classmethod
    def list_properties(cls):
        """
        List properties available for setting/getting on a sliver (those exposing
        setters)
        :return:
        """
        ret = list()
        exclude_set = {"set_property", "set_properties"}
        for k in dir(cls):
            if k.startswith('set_') and k not in exclude_set:
                ret.append(k[4:])
        return ret

    def set_property(self, prop_name: str, prop_val: Any):
        """
        Lets you set any property exposed via a setter
        :param prop_name:
        :param prop_val:
        :return:
        """
        try:
            return self.__getattribute__('set_' + prop_name)(prop_val)
        except AttributeError:
            raise RuntimeError(f'Unable to set property {prop_name} of the sliver - no such property available')

    def get_property(self, prop_name: str):
        """
        Lets you get a property that is exposed via getter method
        :param prop_name:
        :return:
        """
        try:
            return self.__getattribute__('get_' + prop_name)()
        except AttributeError:
            raise RuntimeError(f'Unable to get property {prop_name} of the sliver - no such property available')

    def __repr__(self):
        exclude_set = {"get_property"}
        print_set = list()
        for k in dir(self):
            if k.startswith('get_') and k not in exclude_set:
                print_set.append(k[4:])
        print_set.sort()
        print_vals = dict()
        for p in print_set:
            print_vals[p] = self.get_property(p)
        return str(print_vals)

    def __str__(self):
        return self.__repr__()


class BaseSliverWithDelegation(BaseSliver):
    """
    Intended to be used in advertisements
    """
    def __init__(self):
        super().__init__()
        self.capacity_delegations = None
        self.label_delegations = None

    def set_capacity_delegations(self, cap_del) -> None:
        self.capacity_delegations = cap_del

    def set_label_delegations(self, lab_del) -> None:
        self.label_delegations = lab_del



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ARM (Aggregate Resource Model) functionality
"""

from abc import ABCMeta, abstractmethod
from typing import List

from ..abc_property_graph import ABCPropertyGraph
from fim.slivers.delegations import DelegationType


class ABCARMMixin(metaclass=ABCMeta):
    """
    Interface for an ARM Mixin on top of a property graph
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'generate_adms') and
                callable(subclass.generate_adms) or NotImplemented)

    @abstractmethod
    def generate_adms(self) -> List[ABCPropertyGraph]:
        """
        Partition an already loaded ARM model into multiple ADMs, return the GUIDs
        of the ADM partitions
        :return:
        """

    @abstractmethod
    def get_delegations(self, *, node_id: str, delegation_type: DelegationType) -> List:
        """
        Get Label or Capacity delegation from an ARM as a Python object. They are represented
        as lists of dictionaries, with fields like ipv4, vlan or vlan-pool. May also include fields
        to denote specific delegations.
        :param node_id:
        :param delegation_type:
        :return:
        """


from fim.user.topology import ExperimentTopology
from fim.user.node import Node

if __name__ == "__main__":

    t = ExperimentTopology()
    n1 = t.add_node(name='node1', site='RENC', cpu_cores=1, image_ref="https://some.image")
    print(n1)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import Any, List, Dict
import uuid

from .model_element import ModelElement, ElementType
from .interface import Interface
from ..slivers.switch_fabric import SwitchFabricSliver, SFType, SFLayer
from ..slivers.interface_info import InterfaceType
from ..graph.abc_property_graph import ABCPropertyGraph


class SwitchFabric(ModelElement):
    """
    SwitchFabric class
    """
    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 layer: SFLayer = SFLayer.L2, parent_node_id: str = None):
        """
        Create a new switch fabric. Don't call this yourself, call Node.add_switch_fabric().
        NodeID is mandatory in substrate models.
        :param name:
        :param node_id:
        :param topo:
        :param layer: layer of the switch fabric (defaults to L2)
        :param parent_node_id: id of parent node for new nodes
        :param etype:
        """
        assert name is not None
        assert topo is not None
        # cant use isinstance as it would create circular import dependencies
        if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and node_id is None:
            raise RuntimeError("When adding switch fabrics to substrate topology nodes you must specify static Node ID")
        if node_id is None:
            node_id = str(uuid.uuid4())
        super().__init__(name=name, node_id=node_id, topo=topo)
        if etype == ElementType.NEW:
            if parent_node_id is None:
                raise RuntimeError("For new switch fabrics parent node id must be specified")
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and node_id is None:
                raise RuntimeError("When adding switch fabrics to substrate topology nodes "
                                   "you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            sfsliver = SwitchFabricSliver()
            sfsliver.node_id = node_id
            sfsliver.set_resource_name(name)
            sfsliver.set_resource_type(SFType.SwitchFabric)
            sfsliver.set_layer(layer)
            self.topo.graph_model.add_switch_fabric_sliver(parent_node_id=parent_node_id, switch_fabric=sfsliver)
        else:
            # check that this node exists
            existing_node_id = self.topo.graph_model.find_node_by_name(node_name=name,
                                                                       label=str(ABCPropertyGraph.CLASS_SwitchFabric))
            if node_id is not None and existing_node_id != node_id:
                raise RuntimeError("Existing node id does not match provided. "
                                   "In general you shouldn't need to specify node id for existing nodes.")
            self.node_id = node_id

    def add_interface(self, *, name: str, node_id: str = None, itype: InterfaceType = InterfaceType.TrunkPort):
        """
        Add an interface to node (mostly needed in substrate topologies)
        :param name:
        :param node_id:
        :param itype: interface type e.g. TrunkPort, AccessPort or VINT
        :return:
        """
        iff = Interface(name=name, node_id=node_id, parent_node_id=self.node_id,
                        etype=ElementType.NEW, topo=self.topo, itype=itype)
        return iff

    def remove_interface(self, *, name: str) -> None:
        """
        Remove an interface from the switch fabric, disconnect from links. Remove links
        if they have nothing else connecting to them.
        :param name:
        :return:
        """
        assert name is not None

        node_id = self.topo.graph_model.find_connection_point_by_name(parent_node_id=self.node_id,
                                                                      iname=name)
        self.topo.graph_model.remove_cp_and_links(node_id=node_id)

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a switch fabric property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        sf_sliver = self.topo.graph_model.switch_fabric_sliver_from_graph_properties_dict(node_properties)
        return sf_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a switch fabric property
        :param pname:
        :param pval:
        :return:
        """
        sf_sliver = SwitchFabricSliver()
        sf_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.switch_fabric_sliver_to_graph_properties_dict(sf_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the switch fabtic
        :param kwargs:
        :return:
        """
        sf_sliver = SwitchFabricSliver()
        sf_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.switch_fabric_sliver_to_graph_properties_dict(sf_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> List[str]:
        return SwitchFabricSliver.list_properties()

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of switch fabric by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __list_interfaces(self) -> Dict[str, Interface]:
        """
        List all interfaces of the topology as a dictionary
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_sf_connection_points(parent_node_id=self.node_id)
        # Could consider using frozendict or other immutable idioms
        ret = dict()
        for nid in node_id_list:
            c = self.__get_interface_by_id(nid)
            ret[c.name] = c
        return ret

    def __getattr__(self, item):
        """
        Special handling for attributes like 'components' and 'interfaces' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """
        if item == 'interfaces':
            return self.__list_interfaces()

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        sf_sliver = self.topo.graph_model.switch_fabric_sliver_from_graph_properties_dict(node_properties)
        return sf_sliver.__repr__()

    def __str__(self):
        return self.__repr__()




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Implementation of ARM (Aggregate Resource Model) functionality
"""
import json
import logging
import uuid
from typing import List, Dict
from recordclass import recordclass


from ..neo4j_property_graph import Neo4jPropertyGraph
from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from .abc_arm import ABCARMMixin
from ..typed_tuples import Label, Capacity, LabelOrCapacityException

from fim.slivers.delegations import DelegationType, Pools, Delegations, Delegation


class Neo4jARMGraph(Neo4jPropertyGraph, ABCARMMixin):
    """
    Implementation of ARM management
    """
    POOL_TYPE_TO_CLASS = {DelegationType.LABEL: (ABCPropertyGraph.PROP_LABEL_DELEGATIONS,
                                                 ABCPropertyGraph.FIELD_LABEL_POOL, Label),
                          DelegationType.CAPACITY: (ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS,
                                                    ABCPropertyGraph.FIELD_CAPACITY_POOL, Capacity)}

    DEFAULT_DELEGATION = "default"
    LabelsAndProperties = recordclass('LabelsAndProperties', ['labels', 'properties'])
    DelegationInfo = recordclass('DelegationInfo', ['graph_id',
                                                    'graph',
                                                    'keep_nodes',
                                                    'remove_nodes'])

    def __init__(self, *, graph: Neo4jPropertyGraph):
        """
        Initialize ARM - supply an implementation of a graph and a graph id
        :param graph:
        :param graph:
        """
        super().__init__(graph_id=graph.graph_id, importer=graph.importer)
        self.label_pools = Pools(atype=DelegationType.LABEL)
        self.capacity_pools = Pools(atype=DelegationType.CAPACITY)
        self.label_delegations = Delegations(atype=DelegationType.LABEL)
        self.capacity_delegations = Delegations(atype=DelegationType.CAPACITY)
        self.pools = {DelegationType.LABEL: self.label_pools,
                      DelegationType.CAPACITY: self.capacity_pools}
        self.delegations = {DelegationType.LABEL: self.label_delegations,
                            DelegationType.CAPACITY: self.capacity_delegations}
        self.log = logging.getLogger(__name__)
        self.nodes = dict()
        self.node_ids = None

    def _find_pool_definition(self, *, delegation: Dict, pool_type: DelegationType) -> (str, str):
        """
        Inspect a single delegation to find pool definition and delegation fields, return as a Tuple
        <pool_id, delegation_id>. When delegation is not specified, delegation_id is set to "default".
        :param delegation: single dictionary of a delegation
        :param pool_type: type of a pool (label or capacity)
        :return:
        """
        _, pool_field, pool_cls = self.POOL_TYPE_TO_CLASS[pool_type]
        self.log.debug(f"Looking for pool definition in {delegation} of type {pool_type.name}")
        if pool_field in delegation.keys():
            # parse the delegation pool
            pool = pool_cls(atype=pool_field, aval=delegation[pool_field])
            if self.FIELD_DELEGATION in delegation.keys():
                delegation_id = delegation[self.FIELD_DELEGATION]
            else:
                delegation_id = self.DEFAULT_DELEGATION
            return pool.get_val(), delegation_id
        return None, None

    def _find_pool_mentions(self, *, delegation: Dict, pool_type: DelegationType) -> List[str]:
        """
        Inspect a single delegation to find pool mentions and return a list of all mentioned pool_ids
        or empty list
        :param delegation: single dictionary of a delegation
        :param pool_type:
        :return:
        """
        self.log.debug(f"Looking for pool mention in {delegation} of type {pool_type.name}")
        if Neo4jPropertyGraph.FIELD_POOL in delegation.keys():
            # return the list
            mention = delegation[Neo4jPropertyGraph.FIELD_POOL]
            if isinstance(mention, list):
                return mention
            return [mention]
        return []

    def _process_delegation_no_pools(self, *, delegation: Dict, dele_type: DelegationType, node_id: str) -> Delegation:
        """
        process a delegation with no pools, get a delegation id from delegation dictionary or
        assign 'default' and return a fully formed Delegation object
        :param delegation:
        :param dele_type:
        :return: delegation id
        """
        # extract the delegation id
        if self.FIELD_DELEGATION in delegation.keys():
            delegation_id = delegation[self.FIELD_DELEGATION]
        else:
            delegation_id = self.DEFAULT_DELEGATION
        delegation_obj = Delegation(atype=dele_type, defined_on=node_id, delegation_id=delegation_id)
        delegation_obj.set_details(dele_dict=delegation)
        return delegation_obj

    def _process_single_delegation(self, *, delegation: Dict, dele_type: DelegationType, node_id: str):
        """
        Process a single delegation dictionary for pool definitions, pool mentions or individual delegations.
        Update pools and delegations fields of this object
        :param delegation:
        :param dele_type:
        :param node_id
        :return:
        """
        # check for pool definitions
        pool_id, del_id = self._find_pool_definition(delegation=delegation,
                                                     pool_type=dele_type)
        self.log.debug(f"Single: Decoded pool {pool_id}, delegation {del_id}")
        with_pool = False

        if pool_id is not None:
            # if the pool is not defined, create a new one, if it is defined augment the definition
            pool = self.pools[dele_type].get_pool_by_id(pool_id=pool_id)
            pool.set_delegation_id(delegation_id=del_id)
            pool.set_defined_on(node_id=node_id)
            pool.set_pool_details(pool_dict=delegation)
            with_pool = True

        # check for pool mentions
        pool_mentions = self._find_pool_mentions(delegation=delegation, pool_type=dele_type)
        for pool_id in pool_mentions:
            # if this pool already defined, add node id to list for which it is defined
            pool = self.pools[dele_type].get_pool_by_id(pool_id=pool_id)
            pool.add_defined_for(node_ids=node_id)
            with_pool = True

        # check for plain delegation (without a pool)
        if not with_pool:
            delegation_obj = self._process_delegation_no_pools(delegation=delegation,
                                                               dele_type=dele_type,
                                                               node_id=node_id)
            self.delegations[dele_type].add_delegation(delegation=delegation_obj)

    def _find_delegations_in_properties(self, *, props: Dict, pool_type: DelegationType, node_id: str) -> None:
        """
        Check if a delegation property of a node ( label or capacity) contains a
        pool (label or capacity) declaration, return a list of (pool_id, delegation_id, JSON delegation)
        tuples or empty list. Also process single delegations not included in pool definitions
        :param props: all properties of a node
        :param pool_type
        :param node_id
        :return:
        """
        if pool_type not in self.POOL_TYPE_TO_CLASS.keys():
            raise LabelOrCapacityException(f"Invalid pool type {pool_type}")

        delegation_prop_name, _, _ = self.POOL_TYPE_TO_CLASS[pool_type]
        self.log.debug(f"Checking properties {props} for {delegation_prop_name} pool definitions")
        if delegation_prop_name in props.keys():
            # get the list of delegations as a JSON string
            # note that Neo4j returns string 'None' if property is not filled, but present
            if props[delegation_prop_name] == self.NEO4j_NONE:
                return

            self.log.debug(f"Decoding {delegation_prop_name} delegations {props[delegation_prop_name]} on node {node_id}")
            delegations = json.loads(props[delegation_prop_name])

            if not isinstance(delegations, list):
                if not isinstance(delegations, dict):
                    raise LabelOrCapacityException(f"Delegation {delegations} must be a list "
                                                   "of dictionaries or a dictionary")
                # just a single dictionary
                self._process_single_delegation(delegation=delegations,
                                                dele_type=pool_type, node_id=node_id)
            else:
                # a list of dictionaries
                for delegation in delegations:
                    self._process_single_delegation(delegation=delegation,
                                                    dele_type=pool_type, node_id=node_id)

    def _catalog_delegations(self) -> None:
        """
        Locate all nodes with delegations in them, taking into account pools.
        :param graph_id:
        :return:
        """

        # collect information about pools
        for node_id in self.node_ids:
            self.nodes[node_id].labels, self.nodes[node_id].properties = self.get_node_properties(node_id=node_id)
            self.log.debug(f"Processing node {node_id} for {self.PROP_LABEL_DELEGATIONS} "
                           f"and {self.PROP_CAPACITY_DELEGATIONS}")
            # if a node doesn't have delegations we skip it
            if self.PROP_LABEL_DELEGATIONS not in self.nodes[node_id].properties.keys() and \
                    self.PROP_CAPACITY_DELEGATIONS not in self.nodes[node_id].properties.keys():
                continue
            # look for label and capacity pool definitions and mentions
            self._find_delegations_in_properties(props=self.nodes[node_id].properties,
                                                 pool_type=DelegationType.LABEL, node_id=node_id)

            self._find_delegations_in_properties(props=self.nodes[node_id].properties,
                                                 pool_type=DelegationType.CAPACITY, node_id=node_id)
            # also look for delegations without pools

        # build indices on pools so we can find them from a delegation id
        for _, pools in self.pools.items():
            pools.build_index_by_delegation_id()

    def _update_delegation_properties(self, *, del_id: str, del_type: DelegationType,
                                      graph: Neo4jPropertyGraph,
                                      node_ids: List[str]):
        """
        Update capacity or label delegations on graph nodes based on previously collection
        delegations information. Use the information about kept nodes to speed things up.
        :param del_id:
        :param del_type:
        :param graph:
        :param node_ids:
        :return:
        """
        assert node_ids is not None
        assert graph is not None
        assert del_id is not None

        delegation_prop_name, _, _ = self.POOL_TYPE_TO_CLASS[del_type]
        collected_delegations = self.delegations[del_type]
        delegations_list = collected_delegations.get_by_delegation_id(delegation_id=del_id)
        collected_pools = self.pools[del_type]
        pools_list = collected_pools.get_pools_by_delegation_id(delegation_id=del_id)

        # first fill in a temporary dict with delegation info, then write it into properties
        new_delegations = dict()
        # collect pool mentions separately, then merge
        new_pool_mentions = dict()
        # walk through all node_ids and check delegations and pools
        for node_id in node_ids:
            # each entry matches a node id and is a list of dictionaries
            new_delegations[node_id] = list()

            if delegations_list is not None:
                # find if there is a delegation on this node
                for delegation in delegations_list:
                    if delegation.on_ == node_id:
                        new_delegations[node_id].append(delegation.get_details())

            if pools_list is not None:
                for pool in pools_list:
                    if pool.on_ == node_id:
                        # find if there was a pool definition on this node
                        new_delegations[node_id].append(pool.get_pool_details())
                        # pool definition includes a pool mention
                        if new_pool_mentions.get(node_id, None) is None:
                            new_pool_mentions[node_id] = list()
                        new_pool_mentions[node_id].append(pool.pool_id)
                    elif node_id in pool.for_:
                        # find if there was a pool mention on this node
                        # add it to 'pool' (if exists, create otherwise)
                        if new_pool_mentions.get(node_id, None) is None:
                            new_pool_mentions[node_id] = list()
                        new_pool_mentions[node_id].append(pool.pool_id)

        # merge pool mentions
        for node_id, pool in new_pool_mentions.items():
            new_delegations[node_id].append({Neo4jPropertyGraph.FIELD_POOL: pool})

        # write appropriate property name into the node on appropriate graph
        for node_id in node_ids:
            node_delegation_info = new_delegations[node_id]
            if node_delegation_info is not None and len(node_delegation_info) != 0:
                node_delegation_str = json.dumps(node_delegation_info)
                graph.update_node_property(node_id=node_id,
                                           prop_name=delegation_prop_name,
                                           prop_val=node_delegation_str)

    def generate_adms(self) -> Dict:
        """
        Generate delegation models from the current ARM graph and return
        the dictionary delegation_id: Neo4jPropertyGraph for generated ADMs
        :return:
        """
        self.node_ids = self.list_all_node_ids()
        if len(self.node_ids) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              node_id=None, msg="Unable to list nodes of ARM graph")
        # build a dictionary to hold node labels and properties
        self.nodes = {k: self.LabelsAndProperties(None, None) for k in self.node_ids}

        # collect information on delegated pools and individual delegations
        self._catalog_delegations()

        # number of distinct delegation keys across label and capacity pools and plain delegations
        # is the number of ADMs we will generate
        unique_delegation_ids = self.label_pools.\
            get_delegation_ids().union(self.capacity_pools.get_delegation_ids(),
                                       self.label_delegations.get_delegation_ids(),
                                       self.capacity_delegations.get_delegation_ids())

        # generate unique graph ids and empty node sets for delegation graphs
        # remember these tuples and the fields are not mutable
        delegations_info = {del_id: self.DelegationInfo(graph_id=str(uuid.uuid4()),
                                                        graph=None,
                                                        keep_nodes=set(),
                                                        remove_nodes=set())
                            for del_id in unique_delegation_ids}

        node_ids_to_delegations = {node_id: list() for node_id in self.node_ids}
        for del_id in unique_delegation_ids:
            # build up lists of node ids that definitely belong to each delegation
            # as a union of all .for_ fields on pools and on all delegations
            delegations_info[del_id].keep_nodes.update(
                self.label_pools.get_node_ids(delegation_id=del_id),
                self.capacity_pools.get_node_ids(delegation_id=del_id),
                self.label_delegations.get_node_ids(delegation_id=del_id),
                self.capacity_delegations.get_node_ids(delegation_id=del_id)
            )
            # also build a dictionary from node_id to a list of delegation ids
            for node_id in delegations_info[del_id].keep_nodes:
                node_ids_to_delegations[node_id].append(del_id)

            # clone the original ARM graph into ADM graphs for each new graph ID
            delegations_info[del_id].graph = self.clone_graph(new_graph_id=delegations_info[del_id].graph_id)

        # in addition to definite keep nodes, we should also keep nodes that
        # a) lie on shortest paths between two connection points we are keeping (to include the links)
        # b) switch fabrics that connect to connection points
        # c) network nodes that have switch fabrics
        # Determine nodes to delete by doing a difference between all nodes and keep nodes
        # Remove nodes
        # Annotate keep nodes with proper delegation information

        for del_id in unique_delegation_ids:
            # get a list of connection points we are keeping
            keep_cps = set()
            for node_id in delegations_info[del_id].keep_nodes:
                if 'ConnectionPoint' in self.nodes[node_id].labels:
                    keep_cps.add(node_id)

            # trace from connection points over links to other connection points
            # remember connection points
            new_cps = set()
            for cp in keep_cps:
                cp_neighbors = self.get_first_and_second_neighbor(node_id=cp,
                                                                  rel1="connects",
                                                                  node1_label="Link",
                                                                  rel2="connects",
                                                                  node2_label="ConnectionPoint")
                for pair in cp_neighbors:
                    # pair is a 2-element list
                    delegations_info[del_id].keep_nodes.update(pair)
                    new_cps.add(pair[1])

            keep_cps.update(new_cps)
            # from all keep cps (original and added), find switch fabrics and components or
            # network nodes
            for cp in keep_cps:
                cp_neighbors = self.get_first_and_second_neighbor(node_id=cp,
                                                                  rel1="connects",
                                                                  node1_label="SwitchFabric",
                                                                  rel2="has",
                                                                  node2_label="NetworkNode")
                for pair in cp_neighbors:
                    delegations_info[del_id].keep_nodes.update(pair)

                cp_neighbors = self.get_first_and_second_neighbor(node_id=cp,
                                                                  rel1="connects",
                                                                  node1_label="SwitchFabric",
                                                                  rel2="has",
                                                                  node2_label="Component")
                for pair in cp_neighbors:
                    delegations_info[del_id].keep_nodes.update(pair)

            delegations_info[del_id].remove_nodes = set(self.node_ids)
            delegations_info[del_id].remove_nodes.difference_update(delegations_info[del_id].keep_nodes)
            for node_id in delegations_info[del_id].remove_nodes:
                delegations_info[del_id].graph.delete_node(node_id=node_id)

            # go through capacity delegations under this delegation id, replace annotations on kept nodes
            # of approprate graph
            self._update_delegation_properties(del_id=del_id,
                                               del_type=DelegationType.LABEL,
                                               graph=delegations_info[del_id].graph,
                                               node_ids=delegations_info[del_id].keep_nodes)
            self._update_delegation_properties(del_id=del_id,
                                               del_type=DelegationType.CAPACITY,
                                               graph=delegations_info[del_id].graph,
                                               node_ids=delegations_info[del_id].keep_nodes)

        return {k: delegations_info[k].graph for k in unique_delegation_ids}

    def get_delegations(self, *, node_id: str, delegation_type: DelegationType) -> List:
        """
        Get Label or Capacity delegations of a given node. They are represented
        as lists of dictionaries. Dictionaries may include delegation and pool
        designators. Returns None if delegation property is not defined.
        :param node_id:
        :param delegation_type:
        :return:
        """
        assert node_id is not None
        prop_field_name, _, _ = self.POOL_TYPE_TO_CLASS[delegation_type]
        return self.get_node_json_property_as_object(node_id=node_id, prop_name=prop_field_name)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

"""
Abstract Base class representing operations on a property graph of resources.
Could be a delegation, a broker view of resources or a slice.
"""
import json
import uuid
import networkx as nx
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Set


class ABCPropertyGraph(ABC):
    """
    Abstract Base class representing operations on a property graph of resources.
    Could be a delegation, a broker view of resources or a slice.
    NOTE: this makes a very fundamental assumption that all graphs are sharing
    a common store and nodes of different graphs can be connected to each other.
    Graphs are distinguished by GraphID property on each node.
    """

    FIELD_POOL = "pool"
    FIELD_CAPACITY_POOL = "capacity_pool"
    FIELD_LABEL_POOL = "label_pool"
    FIELD_DELEGATION = "delegation"
    PROP_CAPACITY_DELEGATIONS = "CapacityDelegations"
    PROP_LABEL_DELEGATIONS = "LabelDelegations"
    PROP_CAPACITIES = "Capacities"
    PROP_LABELS = "Labels"
    JSON_PROPERTY_NAMES = [PROP_LABELS, PROP_CAPACITIES, PROP_LABEL_DELEGATIONS, PROP_CAPACITY_DELEGATIONS]
    GRAPH_ID = 'GraphID'
    NODE_ID = 'NodeID'
    PROP_NAME = 'Name'
    PROP_CLASS = 'Class'
    PROP_TYPE = 'Type'
    PROP_MODEL = 'Model'
    PROP_LAYER = 'Layer'
    PROP_TECHNOLOGY = 'Technology'
    PROP_SITE = 'Site'
    PROP_IMAGE_REF = 'ImageRef'
    PROP_MGMT_IP = 'MgmtIp'
    PROP_ALLOCATION_CONSTRAINTS = 'AllocationConstraints'
    PROP_SERVICE_ENDPOINT = 'ServiceEndpoint'
    PROP_DETAILS = 'Details'

    CLASS_NetworkNode = 'NetworkNode'
    CLASS_Component = 'Component'
    CLASS_SwitchFabric = 'SwitchFabric'
    CLASS_ConnectionPoint = 'ConnectionPoint'
    CLASS_Link = 'Link'
    CLASS_CompositeLink = 'CompositeLink'
    CLASS_CompositeNode = 'CompositeNode'
    CLASS_MeasurementPoint = 'MeasurementPoint'

    REL_HAS = 'has'
    REL_CONNECTS = 'connects'
    REL_DEPENDS = 'depends'
    REL_ADAPTS = 'adapts'
    REL_PEERS = 'peers'

    @abstractmethod
    def __init__(self, *, graph_id: str, importer):
        """
        New graph with an id
        :param graph_id:
        """
        assert graph_id is not None
        assert importer is not None
        self.graph_id = graph_id
        self.importer = importer

    def get_graph_id(self):
        return self.graph_id

    @abstractmethod
    def validate_graph(self) -> None:
        """
        validate graph according to a built-in set of rules
        :return: - None,
        """

    @abstractmethod
    def delete_graph(self) -> None:
        """
        delete a graph from the database
        :return: - None
        """

    @abstractmethod
    def get_node_properties(self, *, node_id: str) -> (List[str], Dict[str, Any]):
        """
        return all labels and properties of a node node_id in graph graph_id
        :param node_id:
        :return: dictionary[string, Any]
        """

    @abstractmethod
    def get_node_json_property_as_object(self, *, node_id: str, prop_name: str) -> Any:
        """
        Return node property as a python object (applies to JSON encoded properties), will
        fail on other property values.
        :param node_id:
        :param prop_name:
        :return:
        """

    @abstractmethod
    def get_link_properties(self, *, node_a: str, node_b: str) -> (str, Dict[str, Any]):
        """
        return a tuple of link kind and all properties of a link between two nodes node_a and node_b
        :param node_a:
        :param node_b:
        :param kind: kind of link/edge
        :return: (str, dictionary[string, Any])
        """

    @abstractmethod
    def update_node_property(self, *, node_id: str, prop_name: str,
                             prop_val: Any) -> None:
        """
        update a selected property of a node
        :param node_id:
        :param prop_name:
        :param prop_val:
        :return:
        """

    @abstractmethod
    def update_nodes_property(self, *, prop_name: str, prop_val: Any) -> None:
        """
        update a selected property on all nodes of the graph
        :param prop_name:
        :param prop_val:
        :return:
        """

    @abstractmethod
    def update_node_properties(self, *, node_id: str, props: Dict[str, Any]) -> None:
        """
        update multiple properties
        :param node_id:
        :param props:
        :return:
        """

    @abstractmethod
    def update_link_property(self, *, node_a: str, node_b: str, kind: str,
                             prop_name: str, prop_val: Any) -> None:
        """
        update a link property for a link between node_a and node_b
        :param node_a:
        :param node_b:
        :param kind: - link/relationship type
        :param prop_name:
        :param prop_val:
        :return:
        """

    @abstractmethod
    def update_link_properties(self, *, node_a: str, node_b: str, kind: str,
                               props: Dict[str, Any]) -> None:
        """
        update multiple properties on a link between node_a and node_b
        :param node_a:
        :param node_b:
        :param kind: - link/relationship type
        :param props:
        :return:
        """

    @abstractmethod
    def list_all_node_ids(self) -> List[str]:
        """
        list all NodeID properties of a given graph
        :return:
        """

    @abstractmethod
    def serialize_graph(self) -> str:
        """
        Serialize a given graph into a string
        :return:
        """

    def clone_graph(self, *, new_graph_id: str):
        """
        Clone a graph to a new graph_id by serializing/deserializing it
        :param new_graph_id:
        :return:
        """
        assert new_graph_id is not None
        graph_string = self.serialize_graph()
        if graph_string is None:
            raise PropertyGraphQueryException(msg=f"Unable to find graph with id {self.graph_id} for cloning")
        return self.importer.import_graph_from_string(graph_string=graph_string,
                                                      graph_id=new_graph_id)

    @abstractmethod
    def graph_exists(self) -> bool:
        """
        Does the graph with this ID exist?
        :return:
        """

    @abstractmethod
    def get_nodes_on_shortest_path(self, *, node_a: str, node_z: str, rel: str = None) -> List:
        """
        Get a list of node ids that lie on a shortest path between two nodes. Return empty
        list if no path can be found. Optionally specify the type of relationship that path
        should consist of.
        :param node_a:
        :param node_z:
        :param rel:
        :return:
        """

    @abstractmethod
    def get_first_neighbor(self, *, node_id: str, rel: str, node_label: str) -> List:
        """
        Return a list of ids of nodes of this label related via relationship. List may be empty.
        :param node_id:
        :param rel:
        :param node_label:
        :return:
        """

    @abstractmethod
    def get_first_and_second_neighbor(self, *, node_id: str, rel1: str, node1_label: str,
                                      rel2: str, node2_label: str) -> List:
        """
        Return a list of 2-member lists of node ids related to this node via two specified relationships.
        List may be empty.
        :param node_id:
        :param rel1:
        :param node1_label:
        :param rel2:
        :param node2_label:
        :return:
        """

    @abstractmethod
    def node_exists(self, *, node_id: str, label: str):
        """
        Check if this node exists
        :param node_id:
        :param label:
        :return:
        """

    @abstractmethod
    def add_node(self, *, node_id: str, label: str, props: Dict[str, Any]) -> None:
        """
        Add a new node with specified
        :param node_id:
        :param label: class or label of this node
        :param props: initial set of properties
        :return:
        """

    @abstractmethod
    def add_link(self, *, node_a: str, rel: str, node_b: str, props: Dict[str, Any] = None) -> None:
        """
        Add a link of specified type (rel) between the two nodes. Properties can be added
        at the same time.
        :param node_a:
        :param rel:
        :param node_b:
        :param props:
        :return:
        """

    @abstractmethod
    def delete_node(self, *, node_id: str):
        """
        Delete node from a graph (incident edges automatically deleted)
        :param node_id:
        :return:
        """

    @abstractmethod
    def find_matching_nodes(self, *, other_graph) -> Set:
        """
        Return a set of node ids that match between the two graphs
        :param other_graph:
        :return:
        """

    @abstractmethod
    def merge_nodes(self, node_id: str, other_graph, merge_properties=None):
        """
        Merge two nodes of the same id belonging to two graphs. Optionally
        specify merging behavior for individual properties
        :param node_id:
        :param other_graph:
        :param merge_properties:
        :return:
        """

    def __repr__(self):
        return f"Graph with id {self.graph_id}"

    def _validate_json_property(self, node_id: str, prop_name: str, strict: bool = False) -> str:
        """
        Validate that JSON in a particular node in a particular property is valid or throw
        a JSONDecodeError exception. Strict set to true causes exception if property is not found.
        Default strict is set to False. If property is not a valid JSON it's value is returned.
        :param node_id:
        :param prop_name:
        :param strict
        :return:
        """
        assert node_id is not None
        assert prop_name is not None
        _, props = self.get_node_properties(node_id=node_id)
        if prop_name not in props.keys():
            if strict:
                # if property is not there, raise exception
                raise PropertyGraphImportException(graph_id=self.graph_id, node_id=node_id,
                                                   msg=f"Unable to find property {prop_name} on a node")
            else:
                # if property is not there, just return
                return None
        # try loading it as JSON. Exception may be thrown
        if props[prop_name] is not None and props[prop_name] != "None":
            try:
                json.loads(props[prop_name])
            except json.decoder.JSONDecodeError:
                return props[prop_name]
        else:
            return None

    def _validate_all_json_properties(self) -> None:
        """
        Validate all expected JSON properties of all nodes in a given graph or raise an exception
        :return:
        """
        nodes = self.list_all_node_ids()
        if len(nodes) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id,
                                              node_id=None, msg="Unable to list nodes of graph")
        for node in nodes:
            for prop in self.JSON_PROPERTY_NAMES:
                prop_val = self._validate_json_property(node_id=node, prop_name=prop)
                if prop_val is not None:
                    raise PropertyGraphImportException(graph_id=self.graph_id, node_id=node,
                                                       msg=f"Unable to parse JSON property {prop} with value {prop_val}")


class ABCGraphImporter(ABC):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def import_graph_from_string(self, *, graph_string: str, graph_id: str = None) -> ABCPropertyGraph:
        """
        import a graph from string adding GraphID to each node (new graph_id is generated
        if graph_id is None)
        :param graph_string: - graph represented by a string (e.g. GraphML)
        :param graph_id: - optional id of the graph in the database
        :return: - an instantiation of a property graph
        """

    @abstractmethod
    def import_graph_from_string_direct(self, *, graph_string: str) -> ABCPropertyGraph:
        """
        import a graph from string without any manipulations
        :param graph_string:
        :return:
        """

    def import_graph_from_file(self, *, graph_file: str, graph_id: str = None) -> ABCPropertyGraph:
        """
        import a graph from a file
        :param graph_file: - path to graph file (GraphML)
        :param graph_id: - optional id of the graph in the database
        :return: an instantiation of a property graph
        """
        assert graph_file is not None
        with open(graph_file, 'r') as f:
            graph_string = f.read()

        return self.import_graph_from_string(graph_string=graph_string, graph_id=graph_id)

    @abstractmethod
    def import_graph_from_file_direct(self, *, graph_file: str) -> ABCPropertyGraph:
        """
        import a graph from file without any manipulations
        :param graph_file:
        :return:
        """

    @staticmethod
    def enumerate_graph_nodes(*, graph_file: str, new_graph_file: str, node_id_prop: str = "NodeID") -> None:
        """
        Read in a graph and add a NodeId property to every node assigning a unique GUID (unless present).
        Save into a new file
        :param graph_file: original graph file name
        :param new_graph_file: new file name
        :param node_id_prop: alternative property name for node id (default NodeId)
        :return:
        """
        assert graph_file is not None
        assert new_graph_file is not None

        # read using networkx
        g = nx.read_graphml(graph_file)
        # add node id to every node
        for n in list(g.nodes):
            if (node_id_prop not in g.nodes[n].keys()) or (len(g.nodes[n][node_id_prop]) == 0):
                g.nodes[n][node_id_prop] = str(uuid.uuid4())
        # save to a new file
        nx.write_graphml(g, new_graph_file)

    @staticmethod
    def enumerate_graph_nodes_to_string(*, graph_file: str, node_id_prop: str = "NodeID") -> str:
        """
        Read in a graph and add a NodeId property to every node assigning a unique GUID (unless present).
        Return GraphML string.
        :param graph_file: original graph file name
        :param node_id_prop: alternative property name for node id (default NodeId)
        :return:
        """
        assert graph_file is not None

        # read using networkx
        g = nx.read_graphml(graph_file)
        # add node id to every node
        for n in list(g.nodes):
            if (node_id_prop not in g.nodes[n].keys()) or (len(g.nodes[n][node_id_prop]) == 0):
                g.nodes[n][node_id_prop] = str(uuid.uuid4())
        graph_string = '\n'.join(nx.generate_graphml(g))
        return graph_string

    @abstractmethod
    def delete_graph(self, *, graph_id: str) -> None:
        """
        Delete a single graph with this ID
        :param graph_id:
        :return:
        """

    @abstractmethod
    def delete_all_graphs(self) -> None:
        """
        delete all graphs from the database
        :return:
        """

    @staticmethod
    def get_graph_id(*, graph_file: str) -> str:
        """
        Read graphml file using NetworkX to get GraphID property
        :param graph_file:
        :return:
        """
        assert graph_file is not None
        g = nx.read_graphml(graph_file)

        # check graph_ids on nodes
        graph_ids = set()
        try:
            for n in list(g.nodes):
                graph_ids.add(g.nodes[n][ABCPropertyGraph.GRAPH_ID])
        except KeyError:
            raise PropertyGraphImportException(graph_id=None, msg=f"Graph does not contain GraphID property")

        if len(graph_ids) > 1:
            raise PropertyGraphImportException(graph_id=None, msg=f"Graph contains more than one GraphID: {graph_ids}")
        return graph_ids.pop()


class PropertyGraphException(Exception):
    """
    base exception class for all graph exceptions
    """
    def __init__(self, *, graph_id: str or None, msg: Any = None):
        """
        initialize based on graph_id of the graph in question
        :param graph_id:
        :param msg:
        """
        if msg is None:
            super().__init__(f"Unspecified error importing graph {graph_id}")
        else:
            if graph_id is not None:
                super().__init__(f"{msg} in {graph_id}")
            else:
                super().__init__(msg)
        self.graph_id = graph_id
        self.msg = msg


class PropertyGraphImportException(PropertyGraphException):
    """
    import exception for a property graph
    """
    def __init__(self, *, graph_id: str or None,  msg: str, node_id: str = None):
        if node_id is None:
            super().__init__(graph_id=graph_id, msg=f"Error [{msg}] importing graph")
        else:
            super().__init__(graph_id=graph_id, msg=f"Error [{msg}] in node {node_id} importing graph")


class PropertyGraphQueryException(PropertyGraphException):
    """
    query exception for a property graph
    """
    def __init__(self, *, graph_id: str or None, node_id: str or None, msg: str, node_b: str = None, kind: str = None):
        """
        Query error for node or link
        :param graph_id:
        :param node_id:
        :param msg:
        :param node_b:
        :param kind:
        """
        if node_b is not None and node_id is not None:
            super().__init__(graph_id=graph_id, msg=f"[{msg}] in querying for link {kind} between {node_id} and {node_b}")
        elif node_b is None and node_id is not None:
            super().__init__(graph_id=graph_id, msg=f"[{msg}] in querying node {node_id}")
        else:
            super().__init__(graph_id=graph_id, msg=f"[{msg}]")
        self.node_id = node_id
